import logging
import time
import numpy as np

import rlberry.spaces as spaces
from rlberry.agents import Agent
from rlberry.agents.dynprog.utils import backward_induction
from rlberry.agents.dynprog.utils import backward_induction_in_place
from rlberry.agents.kernel_based.common import map_to_representative


class RSUCBVIAgent(Agent):
    """
    Value iteration with exploration bonuses for continuous-state environments,
    using a online discretization strategy:
    - Build (online) a set of representative states
    - Estimate transtions an rewards on the finite set of representative states
    and actions.

    Criterion: finite-horizon with discount factor gamma.
    If the discount is not 1, only the Q function at h=0 is used.

    The recommended policy after all the episodes is computed without
    exploration bonuses.


    References
    ----------
    Azar, Mohammad Gheshlaghi, Ian Osband, and Rémi Munos.
    "Minimax regret bounds for reinforcement learning."
    Proceedings of the 34th ICML, 2017.

    Strehl, Alexander L., and Michael L. Littman.
    "An analysis of model-based interval estimation for Markov decision
    processes."
     Journal of Computer and System Sciences 74.8 (2008): 1309-1331.

    Kveton, Branislav, and Georgios Theocharous.
    "Kernel-Based Reinforcement Learning on Representative States."
    AAAI, 2012.

    Domingues, O. D., Ménard, P., Pirotta, M., Kaufmann, E., & Valko, M.(2020).
    A kernel-based approach to non-stationary reinforcement learning in metric
    spaces.
    arXiv preprint arXiv:2007.05078.
    """

    name = "RSUCBVI"
    fit_info = ("n_episodes", "episode_rewards")

    def __init__(self, env,
                 n_episodes=1000,
                 gamma=0.95,
                 horizon=None,
                 lp_metric=2,
                 scaling=None,
                 min_dist=0.1,
                 max_repr=1000,
                 bonus_scale_factor=1.0,
                 bonus_type="simplified_bernstein",
                 verbose=1,
                 **kwargs):
        """
        env : Model
            Online model with continuous (Box) state space and discrete actions
        n_episodes : int
            number of episodes
        gamma : double
            Discount factor in [0, 1]. If gamma is 1.0, the problem is set to
            be finite-horizon.
        horizon : int
            Horizon of the objective function. If None and gamma<1, set to
            1/(1-gamma).
        lp_metric: int
            The metric on the state space is the one induced by the p-norm,
            where p = lp_metric. Default = 2, for the Euclidean metric.
        scaling: numpy.ndarray
            Must have the same size as state array, used to scale the states
            before computing the metric.
            If None, set to:
            - (env.observation_space.high - env.observation_space.low) if high
            and low are bounded
            - np.ones(env.observation_space.dim) if high or low are unbounded
        min_dist: double
            Minimum distance between two representative states
        max_repr: int
            Maximum number of representative states.
            If None, it is set to  (sqrt(d)/min_dist)**d, where d
            is the dimension of the state space
        bonus_scale_factor : double
            Constant by which to multiply the exploration bonus, controls
            the level of exploration.
        verbose : int
            Controls the verbosity, if non zero, progress messages are printed.
        bonus_type : string
             Type of exploration bonus. Currently, only "simplified_bernstein"
             is implemented.
        """
        # init base class
        Agent.__init__(self, env, **kwargs)

        self.n_episodes = n_episodes
        self.gamma = gamma
        self.horizon = horizon
        self.lp_metric = lp_metric
        self.min_dist = min_dist
        self.bonus_scale_factor = bonus_scale_factor
        self.verbose = verbose
        self.bonus_type = bonus_type

        # check environment
        assert self.env.is_online()
        assert isinstance(self.env.observation_space, spaces.Box)
        assert isinstance(self.env.action_space, spaces.Discrete)

        # other checks
        assert gamma >= 0 and gamma <= 1.0
        if self.horizon is None:
            assert gamma < 1.0, \
                "If no horizon is given, gamma must be smaller than 1."
            self.horizon = int(np.ceil(1.0 / (1.0 - gamma)))

        # state dimension
        self.state_dim = self.env.observation_space.dim

        # compute scaling, if it is None
        if scaling is None:
            # if high and low are bounded
            if (self.env.observation_space.high == np.inf).sum() == 0 \
                    and (self.env.observation_space.low == -np.inf).sum() == 0:
                scaling = self.env.observation_space.high \
                    - self.env.observation_space.low
                # if high or low are unbounded
            else:
                scaling = np.ones(self.state_dim)
        else:
            assert scaling.ndim == 1
            assert scaling.shape[0] == self.state_dim
        self.scaling = scaling

        # maximum value
        r_range = self.env.reward_range[1] - self.env.reward_range[0]
        if r_range == np.inf:
            logging.warning("{}: Reward range is infinity. ".format(self.name)
                            + "Clipping it to 1.")
            r_range = 1.0

        if self.gamma == 1.0:
            self.v_max = r_range * horizon
        else:
            self.v_max = r_range * (1.0 - np.power(self.gamma, self.horizon)) \
                                                        / (1.0 - self.gamma)

        # number of representative states and number of actions
        if max_repr is None:
            max_repr = int(np.ceil((1.0 * np.sqrt(self.state_dim) /
                                    self.min_dist) ** self.state_dim))
        self.max_repr = max_repr

        # current number of representative states
        self.M = None
        self.A = self.env.action_space.n

        # declaring variables
        self.episode = None  # current episode
        self.representative_states = None  # coordinates of all repr states
        self.N_sa = None  # visits to (s, a)
        self.N_sas = None  # visits to (s, a, s')
        self.S_sa = None  # sum of rewards at (s, a)
        self.B_sa = None  # bonus at (s, a)
        self.Q = None  # Q function
        self.V = None  # V function

        self.Q_policy = None  # Q function for recommended policy

        # initialize
        self.reset()

    def reset(self, **kwargs):
        self.M = 0
        self.representative_states = np.zeros((self.max_repr, self.state_dim))
        self.N_sa = np.zeros((self.max_repr, self.A))
        self.N_sas = np.zeros((self.max_repr, self.A, self.max_repr))
        self.S_sa = np.zeros((self.max_repr, self.A))
        self.B_sa = self.v_max * np.ones((self.max_repr, self.A))

        self.R_hat = np.zeros((self.max_repr, self.A))
        self.P_hat = np.zeros((self.max_repr, self.A, self.max_repr))

        self.V = np.zeros((self.horizon, self.max_repr))
        self.Q = np.zeros((self.horizon, self.max_repr, self.A))
        self.Q_policy = None

        self.episode = 0

        # logging config
        self._last_printed_ep = 0
        self._time_last_log = time.process_time()
        if self.verbose == 1:
            self._log_interval = 60  # in seconds
        elif self.verbose == 2:
            self._log_interval = 30
        elif self.verbose == 3:
            self._log_interval = 15
        elif self.verbose > 3:
            self._log_interval = 5

    def policy(self, state, hh=0, **kwargs):
        assert self.Q_policy is not None
        repr_state = self._map_to_repr(state, False)

        # no discount
        if self.gamma == 1.0:
            return self.Q_policy[hh, repr_state, :].argmax()
        # discounted
        return self.Q_policy[0, repr_state, :].argmax()

    def fit(self, **kwargs):
        info = {}
        self._rewards = np.zeros(self.n_episodes)
        self._cumul_rewards = np.zeros(self.n_episodes)
        for kk in range(self.n_episodes):
            episode_rewards = self._run_episode()
            self._rewards[kk] = episode_rewards
            if kk > 0:
                self._cumul_rewards[kk] = episode_rewards \
                    + self._cumul_rewards[kk - 1]
            self.episode += 1
            #
            self._logging()

        # compute Q function for the recommended policy
        self.Q_policy, _ = backward_induction(self.R_hat[:self.M, :],
                                              self.P_hat[:self.M, :, :self.M],
                                              self.horizon, self.gamma)

        info["n_episodes"] = self.n_episodes
        info["episode_rewards"] = self._rewards
        return info

    def _logging(self):
        if self.verbose > 0:
            t_now = time.process_time()
            time_elapsed = t_now - self._time_last_log
            if time_elapsed >= self._log_interval:
                self._time_last_log = t_now
                print(self._info_to_print())
                self._last_printed_ep = self.episode - 1

    def _info_to_print(self):
        prev_episode = self._last_printed_ep
        episode = self.episode - 1
        reward_per_ep = self._rewards[prev_episode:episode + 1].sum() / \
            max(1, episode - prev_episode)
        time_per_ep = self._log_interval * 1000.0 / \
            max(1, episode - prev_episode)
        time_per_ep = max(0.01, time_per_ep)  # avoid div by zero
        to_print = "[{}] episode = {}/{} ".format(self.name, episode+1,
                                                  self.n_episodes) \
            + "| representative states = {} ".format(self.M) \
            + "| reward/ep = {:0.2f} ".format(reward_per_ep) \
            + "| time/ep = {:0.2f} ms".format(time_per_ep)

        return to_print

    def _map_to_repr(self, state, accept_new_repr=True):
        repr_state = map_to_representative(state,
                                           self.lp_metric,
                                           self.representative_states,
                                           self.M,
                                           self.min_dist,
                                           self.scaling,
                                           accept_new_repr)
        # check if new representative state
        if repr_state == self.M:
            self.M += 1
        return repr_state

    def _update(self, state, action, next_state, reward):
        repr_state = self._map_to_repr(state)
        repr_next_state = self._map_to_repr(next_state)

        self.N_sa[repr_state, action] += 1
        self.N_sas[repr_state, action, repr_next_state] += 1
        self.S_sa[repr_state, action] += reward

        self.R_hat[repr_state, action] = self.S_sa[repr_state, action] \
            / self.N_sa[repr_state, action]
        self.P_hat[repr_state, action, :] = self.N_sas[repr_state, action, :] \
            / self.N_sa[repr_state, action]
        self.B_sa[repr_state, action] = \
            self._compute_bonus(self.N_sa[repr_state, action])

    def _compute_bonus(self, n):
        if self.bonus_type == "simplified_bernstein":
            bonus = self.bonus_scale_factor * np.sqrt(1.0 / n) + self.v_max / n
            bonus = min(bonus, self.v_max)
            return bonus
        else:
            raise NotImplementedError(
                "Error: bonus type {} not implemented".format(self.bonus_type))

    def _get_action(self, state, hh=0):
        assert self.Q is not None
        repr_state = self._map_to_repr(state, False)

        # no discount
        if self.gamma == 1.0:
            return self.Q[hh, repr_state, :].argmax()
        # discounted
        return self.Q[0, repr_state, :].argmax()

    def _run_episode(self):
        # interact for H steps
        episode_rewards = 0
        state = self.env.reset()
        for hh in range(self.horizon):
            action = self._get_action(state, hh)
            next_state, reward, done, _ = self.env.step(action)
            self._update(state, action, next_state, reward)
            state = next_state
            episode_rewards += reward

            if done:
                break

        # run backward induction
        backward_induction_in_place(
            self.Q[:, :self.M, :], self.V[:, :self.M],
            self.R_hat[:self.M, :] + self.B_sa[:self.M, :],
            self.P_hat[:self.M, :, :self.M],
            self.horizon, self.gamma, self.v_max)

        # return sum of rewards collected in the episode
        return episode_rewards



"""
Provide classes for geometric primitives in OpenGL and scenes.
"""


class Scene:
    """
    Class representing a scene, which is a vector of GeometricPrimitive objects
    """

    def __init__(self):
        self.shapes = []

    def add_shape(self, shape):
        self.shapes.append(shape)


class GeometricPrimitive:
    """
    Class representing an OpenGL geometric primitive.

     Primitive type (GL_LINE_LOOP by defaut)

     If using OpenGLRender2D, one of the following:
           POINTS
           LINES
           LINE_STRIP
           LINE_LOOP
           POLYGON
           TRIANGLES
           TRIANGLE_STRIP
           TRIANGLE_FAN
           QUADS
           QUAD_STRIP

    If using PyGameRender2D:
            POLYGON


    TODO: Add support to more pygame shapes,
    see https://www.pygame.org/docs/ref/draw.html
    """

    def __init__(self, primitive_type="GL_LINE_LOOP"):
        # primitive type
        self.type = primitive_type
        # color in RGB
        self.color = (0.25, 0.25, 0.25)
        # list of vertices. each vertex is a tuple with coordinates in space
        self.vertices = []

    def add_vertex(self, vertex):
        self.vertices.append(vertex)

    def set_color(self, color):
        self.color = color



import numpy as np

from rlberry.spaces import Space


class Box(Space):
    """
    Class that represents a space that is a cartesian product in R^n:

    [a_1, b_1] x [a_2, b_2] x ... x [a_n, b_n]

    Attributes
    ----------
    low : numpy.ndarray
         [a_1, ..., a_n]
    high : numpy.ndarray
         [b_1, ..., b_n]
    dim : int
        dimension of the space (n in R^n)
    is_bounded : bool
        true if all the dimensions are bounded
    rng : numpy.random._generator.Generator
        random number generator provided by rlberry.seeding

    Methods
    -------
    sample()
        sample element from the space
    contains(x)
        check if x belongs to the space
    """

    def __init__(self, low, high, dim=None, default_dtype=np.float64):
        """
        Parameters
        -----------
        low : numpy.ndarray or double
            - if low is a double, the parameter dim must be given, and
            low will be set to np.full((dim,), low, dtype=default_dtype)
            - if low is a np.ndarray, default_dtype is ignored
        high : numpy.ndarray or double
            - if high is a double, the parameter dim must be given, and
            high will be set to np.full((dim,), high, dtype=default_dtype)
            - if high is a np.ndarray, default_dtype is ignored
        dim :
            dimension of the space, to be given if low and high are double
        default_dtype :
            default type for values in numpy arrays, used if low and high are
            double
        """
        super(Box, self).__init__()
        assert type(low) is type(high), "low and high must have the same type"

        if type(low) is not np.ndarray:
            assert dim is not None and dim > 0, \
                "dim must be given if low/high are not arrays"
            self.low = np.full((dim,), low, dtype=default_dtype)
            self.high = np.full((dim,), high, dtype=default_dtype)
            self.dim = dim
        else:
            assert low.ndim == high.ndim == 1, \
                "high and low must be 1d"
            assert low.shape == high.shape, \
                "low and high must have the same shape"
            self.low = low.copy()
            self.high = high.copy()
            self.dim = low.shape[0]

        # check which dimensions are bounded
        self.is_bounded = True
        self._boundedness = np.zeros(self.dim, dtype=np.uint8)
        for dd in range(self.dim):
            assert self.low[dd] < self.high[dd], "low < high is required!"
            bounded_below = 1 * (self.low[dd] != -np.inf)
            bounded_above = 1 * (self.high[dd] != np.inf)
            # 0 if not below and not above
            # 1 if not below and     above
            # 2 if     below and not above
            # 3 if     below and     above
            self._boundedness[dd] = 2 * bounded_below + bounded_above
            self.is_bounded = self.is_bounded and (self._boundedness[dd] == 3)

    def sample(self):
        """
        For each dimension i:
        - if bounded above and below:   uniform distribution on [a_i, b_i]
        - if unbounded above:   exponential distribution on [a_i, infty]
        - if unbounded below:   exponential distribution on [-infty, b_i]
        - if unbounded above and below: normal distribution
        """
        xsample = np.zeros(self.dim)
        for dd in range(self.dim):
            # bounded above and below
            if self._boundedness[dd] == 3:
                xsample[dd] = self.rng.uniform(self.low[dd], self.high[dd])
            # unbounded above
            elif self._boundedness[dd] == 2:
                xsample[dd] = self.low[dd] + self.rng.exponential()
            # unbounded below
            elif self._boundedness[dd] == 1:
                xsample[dd] = self.high[dd] - self.rng.exponential()
            # unbounded above and below
            elif self._boundedness[dd] == 0:
                xsample[dd] = self.rng.normal()
        return xsample

    def contains(self, x):
        contain = True
        contain = contain and (type(x) is np.ndarray)
        contain = contain and (x.ndim == 1)
        contain = contain and (x.shape[0] == self.dim)
        if contain:
            for dd in range(self.dim):
                contain = contain and \
                          (x[dd] <= self.high[dd]) and (x[dd] >= self.low[dd])
        return contain

    def __str__(self):
        objstr = "%d-dimensional Box space:\n" % self.dim
        for dd in range(self.dim):
            objstr += "   [%0.2f, %0.2f]\n" % (self.low[dd], self.high[dd])
        return objstr



import numpy as np

import rlberry.spaces as spaces
from rlberry.envs.interface import Model
from rlberry.rendering import Scene, GeometricPrimitive, RenderInterface2D


def projection_to_pball(x, p):
    """
    Solve the problem:
        min_z  ||x-z||_2^2
        s.t.   ||z||_p  <= 1
    for p = 2 or p = np.inf

    If p is not 2 or np.inf, it returns x/norm_p(x) if norm_p(x) > 1

    WARNING: projection_to_pball is not actually a projection for p!=2
    or p=!np.inf
    """
    if np.linalg.norm(x, ord=p) <= 1.0:
        return x

    if p == 2:
        z = x / np.linalg.norm(x, ord=p)
        return z

    if p == np.inf:
        z = np.minimum(1.0, np.maximum(x, -1.0))
        return z

        # below it is not a projection
    return x / np.linalg.norm(x, ord=p)


class PBall(Model):
    """
    Parametric family of environments whose state space is a unit sphere
    according to the p-norm in R^d.

    Note:
        The projection function is only a true projection for
        p in {2, infinity}.

    ----------------------------------------------------------------------
    State space:
        x in R^d: norm_p (x) <= 1

        implemented as rlberry.spaces.Box representing [0, 1]^d
    ----------------------------------------------------------------------
    Action space:
        {u_1, ..., u_m} such that u_i in R^d'  for i = 1, ..., m

        implemented as rlberry.spaces.Discrete(m)
    ----------------------------------------------------------------------
    Reward function (independent of the actions):
        r(x) = sum_{i=1}^n  b_i  max( 0,  1 - norm_p( x - x_i )/c_i )

        requirements:
            c_i >= 0
            b_i in [0, 1]
    ----------------------------------------------------------------------
    Transitions:
        x_{t+1} = A x_t + B u_t + N

        where
            A: square matrix of size d
            B: matrix of size (d, d')
            N: d-dimensional Gaussian noise with zero mean and covariance
            matrix sigma*I
    ----------------------------------------------------------------------
    Initial state:
        d-dimensional Gaussian with mean mu_init and covariance matrix
        sigma_init*I
    ----------------------------------------------------------------------

    Default parameters are provided for a 2D environment, PBall2D
    """

    name = "LP-Ball"

    def __init__(self,
                 p,
                 action_list,
                 reward_amplitudes,
                 reward_smoothness,
                 reward_centers,
                 A,
                 B,
                 sigma,
                 sigma_init,
                 mu_init):
        """
        Parameters
        -----------
        p : int
            parameter of the p-norm
        action_list : list
            list of actions {u_1, ..., u_m}, each action u_i is a
            d'-dimensional array
        reward_amplitudes: list
            list of reward amplitudes: {b_1, ..., b_n}
        reward_smoothness : list
            list of reward smoothness: {c_1, ..., c_n}
        reward_centers : list
            list of reward centers:    {x_1, ..., x_n}
        A : numpy.ndarray
            array A of size (d, d)
        B : numpy.ndarray
            array B of size (d, d')
        sigma : double
            transition noise sigma
        sigma_init : double
            initial state noise sigma_init
        mu_init : numpy.ndarray
            array of size (d,) containing the mean of the initial state
        """
        Model.__init__(self)

        assert p >= 1, "PBall requires p>=1"
        if p not in [2, np.inf]:
            print("WARNING: for p!=2 or p!=np.inf, PBall \
does not make true projections onto the lp ball.")
        self.p = p
        self.d, self.dp = B.shape  # d and d'
        self.m = len(action_list)
        self.action_list = action_list
        self.reward_amplitudes = reward_amplitudes
        self.reward_smoothness = reward_smoothness
        self.reward_centers = reward_centers
        self.A = A
        self.B = B
        self.sigma = sigma
        self.sigma_init = sigma_init
        self.mu_init = mu_init

        # State and action spaces
        low = -1.0 * np.ones(self.d, dtype=np.float64)
        high = np.ones(self.d, dtype=np.float64)
        self.observation_space = spaces.Box(low, high)
        self.action_space = spaces.Discrete(self.m)

        # reward range
        assert len(self.reward_amplitudes) == len(self.reward_smoothness)
        assert len(self.reward_amplitudes) == len(self.reward_centers)
        assert self.reward_amplitudes.max() <= 1.0 and \
            self.reward_amplitudes.min() >= 0.0, \
            "reward amplitudes b_i must be in [0, 1]"
        assert self.reward_smoothness.min() > 0.0, \
            "reward smoothness c_i must be > 0"
        self.reward_range = (0, 1.0)

        #
        self.name = "Lp-Ball"

        # Initalize state
        self.reset()

    def reset(self, state=None):
        if state is not None:
            self.state = state
        else:
            self.state = self.mu_init \
                + self.sigma_init * self.rng.normal(size=self.d)
            # projection to unit ball
        self.state = projection_to_pball(self.state, self.p)
        return self.state.copy()

    def sample(self, state, action):
        assert self.action_space.contains(action)
        assert self.observation_space.contains(state)

        # next state
        action_vec = self.action_list[action]
        next_s = self.A.dot(state) + self.B.dot(action_vec) \
            + self.sigma * self.rng.normal(size=self.d)
        next_s = projection_to_pball(next_s, self.p)

        # done and reward
        done = False
        reward = self.compute_reward_at(state)

        return next_s, reward, done, {}

    def step(self, action):
        next_s, reward, done, info = self.sample(self.state, action)
        self.state = next_s.copy()
        return next_s, reward, done, info

    def compute_reward_at(self, x):
        reward = 0
        for ii, b_ii in enumerate(self.reward_amplitudes):
            c_ii = self.reward_smoothness[ii]
            x_ii = self.reward_centers[ii]
            dist = np.linalg.norm(x - x_ii, ord=self.p)
            reward += b_ii * max(0.0, 1.0 - dist / c_ii)
        return reward

    def get_reward_lipschitz_constant(self):
        ratios = self.reward_amplitudes / self.reward_smoothness
        Lr = ratios.max()
        return Lr

    def get_transitions_lipschitz_constant(self):
        """
        note: considers a fixed action, returns Lipschitz constant
        w.r.t. to states.

        If p!=1, p!=2 or p!=np.inf, returns an upper bound on the induced norm
        """
        if self.p == 1:
            order = np.inf
        else:
            order = self.p / (self.p - 1.0)

        if order in [1, 2]:
            return np.linalg.norm(self.A, ord=order)

        # If p!=1, p!=2 or p!=np.inf, return upper bound on the induced norm.
        return np.power(self.d, 1.0 / self.p) * np.linalg.norm(self.A,
                                                               ord=np.inf)


class PBall2D(PBall, RenderInterface2D):
    def __init__(self,
                 p=2,
                 action_list=[0.05 * np.array([1, 0]),
                              -0.05 * np.array([1, 0]),
                              0.05 * np.array([0, 1]),
                              -0.05 * np.array([0, 1])],
                 reward_amplitudes=np.array([1.0]),
                 reward_smoothness=np.array([0.25]),
                 reward_centers=[np.array([0.75, 0.0])],
                 A=np.eye(2),
                 B=np.eye(2),
                 sigma=0.01,
                 sigma_init=0.001,
                 mu_init=np.array([0.0, 0.0])
                 ):
        # Initialize PBall
        PBall.__init__(self, p, action_list, reward_amplitudes,
                       reward_smoothness,
                       reward_centers,
                       A, B, sigma, sigma_init, mu_init)

        # Render interface
        RenderInterface2D.__init__(self)

        # rendering info
        self.set_clipping_area((-1, 1, -1, 1))
        self.set_refresh_interval(50)  # in milliseconds

    def step(self, action):
        # save state for rendering
        if self.is_render_enabled():
            self.append_state_for_rendering(self.state.copy())
        return PBall.step(self, action)

    #
    # Code for rendering
    #

    def _get_ball_shape(self, xcenter, radius):
        shape = GeometricPrimitive("POLYGON")
        n_points = 200
        theta_vals = np.linspace(0.0, 2 * np.pi, n_points)
        for theta in theta_vals:
            pp = np.array([2.0 * np.cos(theta), 2.0 * np.sin(theta)])
            pp = xcenter + radius * projection_to_pball(pp, self.p)
            # project to the main ball after translation
            pp = projection_to_pball(pp, self.p)
            shape.add_vertex((pp[0], pp[1]))
        return shape

    def get_background(self):
        bg = Scene()

        # ball shape
        contour = self._get_ball_shape(np.zeros(2), 1.0)
        contour.set_color((0.0, 0.0, 0.5))
        bg.add_shape(contour)

        # reward position
        for ii, ampl in enumerate(self.reward_amplitudes):
            contour = self._get_ball_shape(self.reward_centers[ii],
                                           self.reward_smoothness[ii])
            ampl = 1.0 - ampl  # dark violet = more reward
            contour.set_color((0.5, 0.0, 0.5 * (1.0 + ampl)))
            bg.add_shape(contour)

        return bg

    def get_scene(self, state):
        scene = Scene()

        agent = GeometricPrimitive("QUADS")
        agent.set_color((0.75, 0.0, 0.5))
        size = 0.05
        x = state[0]
        y = state[1]
        agent.add_vertex((x - size / 4.0, y - size))
        agent.add_vertex((x + size / 4.0, y - size))
        agent.add_vertex((x + size / 4.0, y + size))
        agent.add_vertex((x - size / 4.0, y + size))

        agent.add_vertex((x - size, y - size / 4.0))
        agent.add_vertex((x + size, y - size / 4.0))
        agent.add_vertex((x + size, y + size / 4.0))
        agent.add_vertex((x - size, y + size / 4.0))

        scene.add_shape(agent)
        return scene


class SimplePBallND(PBall):
    """
    PBall environment in d dimensions with simple dynamics.
    """

    def __init__(self,
                 p=2,
                 dim=2,
                 action_amplitude=0.05,
                 r_smoothness=0.25,
                 sigma=0.01,
                 sigma_init=0.001,
                 mu_init=None
                 ):
        # Action list
        action_list = []
        for dd in range(dim):
            aux = np.zeros(dim)
            aux[dd] = action_amplitude
            action_list.append(aux)
            action_list.append(-1 * aux)

        # Rewards
        reward_amplitudes = np.array([1.0])
        reward_smoothness = np.array([r_smoothness])
        reward_centers = [np.zeros(dim)]
        reward_centers[0][0] = 0.8

        # Transitions
        A = np.eye(dim)
        B = np.eye(dim)

        # Initial position
        if mu_init is None:
            mu_init = np.zeros(dim)

        # Initialize PBall
        PBall.__init__(self, p, action_list, reward_amplitudes,
                       reward_smoothness,
                       reward_centers,
                       A, B, sigma, sigma_init, mu_init)


# if __name__ == '__main__':
#     env = PBall2D(p=5)
#     print(env.get_transitions_lipschitz_constant())
#     print(env.get_reward_lipschitz_constant())

#     env.enable_rendering()

#     for ii in range(100):
#         env.step(1)
#         env.step(3)

#     env.render()



from setuptools import setup, find_packages

packages = find_packages(exclude=['docs', 'notebooks', 'logo'])

install_requires = [
    'numpy>=1.17',
    'pygame',
    'joblib',
    'matplotlib',
    'seaborn',
    'pandas'
]

tests_require = [
    'pytest',
    'pytest-cov',
    'numpy>=1.17',
    'numba',
    'joblib',
    'matplotlib',
    'pandas',
    'seaborn',
    'optuna',
    'pyvirtualdisplay',
]

full_requires = [
    'numba',
    'torch>=1.6.0',
    'optuna',
    'ffmpeg-python',
    'PyOpenGL',
    'PyOpenGL_accelerate',
    'pyvirtualdisplay',
]

extras_require = {
    'full': full_requires,
    'test': tests_require,
    'deploy': ['sphinx', 'sphinx_rtd_theme'],
    'opengl_rendering': ['PyOpenGL', 'PyOpenGL_accelerate'],
    'torch_agents': ['torch>=1.6.0'],
    'hyperparam_optimization': ['optuna'],
    'save_video': ['ffmpeg-python'],
}

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='rlberry',
    version='0.0.1',
    description='An easy-to-use reinforcement learning library',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='rlberry team',
    url='https://github.com/rlberry-py',
    license='MIT',
    packages=packages,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
    zip_safe=False,
)




from rlberry.agents.ppo import PPOAgent
from rlberry.agents.avec import AVECPPOAgent
from rlberry.envs.benchmarks.ball_exploration.ball2d import get_benchmark_env


def test_ppo_agent():
    env = get_benchmark_env(level=1)
    n_episodes = 5
    horizon = 30

    agent = PPOAgent(env,
                     n_episodes=n_episodes,
                     horizon=horizon,
                     gamma=0.99,
                     learning_rate=0.001,
                     eps_clip=0.2,
                     k_epochs=4,
                     verbose=0)
    agent._log_interval = 0
    agent.fit()
    agent.policy(env.observation_space.sample())


def test_ppo_agent_partial_fit():
    env = get_benchmark_env(level=1)
    n_episodes = 10
    horizon = 30

    agent = PPOAgent(env,
                     n_episodes=n_episodes,
                     horizon=horizon,
                     gamma=0.99,
                     learning_rate=0.001,
                     eps_clip=0.2,
                     k_epochs=4,
                     verbose=1)
    agent._log_interval = 0

    agent.partial_fit(0.5)
    agent.policy(env.observation_space.sample())
    assert agent.episode == 5
    agent.partial_fit(0.5)
    assert agent.episode == 10
    agent.policy(env.observation_space.sample())


def test_avec_ppo_agent():
    env = get_benchmark_env(level=1)
    n_episodes = 5
    horizon = 30

    agent = AVECPPOAgent(env,
                         n_episodes=n_episodes,
                         horizon=horizon,
                         gamma=0.99,
                         lr=0.001,
                         eps_clip=0.2,
                         k_epochs=4,
                         verbose=1)
    agent._log_interval = 0
    agent.fit()
    agent.policy(env.observation_space.sample())



import numpy as np
import logging
from rlberry.seeding import seeding
from rlberry.spaces import Space


class Model:
    """
    Base class for an environment model.

    Attributes
    ----------
    name : string
        environment identifier
    observation_space : rlberry.spaces.Space
        observation space
    action_space : rlberry.spaces.Space
        action space
    reward_range : tuple
        tuple (r_min, r_max) containing the minimum and the maximum reward
    rng : numpy.random._generator.Generator
        random number generator provided by rlberry.seeding

    Methods
    -------
    reseed()
        get new random number generator
    reset()
        puts the environment in a default state and returns this state
    step(action)
        returns the outcome of an action
    sample(state, action)
        returns a transition sampled from taking an action in a given state
    is_online()
        returns true if reset() and step() methods are implemented
    is_generative()
        returns true if sample() method is implemented
    """

    name = ""

    def __init__(self):
        self.observation_space: Space = None
        self.action_space: Space = None
        self.reward_range: tuple = (-np.inf, np.inf)
        # random number generator
        self.rng = seeding.get_rng()

    def reseed(self):
        """
        Get new random number generator for the model.
        """
        self.rng = seeding.get_rng()
        self.observation_space.rng = self.rng
        self.action_space.rng = self.rng

    def reset(self):
        """
        Puts the environment in a default state and returns this state.
        """
        raise NotImplementedError("reset() method not implemented.")

    def step(self, action):
        """
        Execute a step. Similar to gym function [1].
        [1] https://gym.openai.com/docs/#environments

        Parameters
        ----------
        action : object
            action to take in the environment

        Returns:
        -------
        observation : object
        reward : float
        done  : bool
        info  : dict
        """
        raise NotImplementedError("step() method not implemented.")

    def sample(self, state, action):
        """
        Execute a step from a state-action pair.

        Parameters
        ----------
        state : object
            state from which to sample
        action : object
            action to take in the environment

        Returns:
        -------
        observation : object
        reward : float
        done  : bool
        info  : dict
        """
        raise NotImplementedError("sample() method not implemented.")

    def is_online(self):
        logging.warning("Checking if Model is\
online calls reset() and step() methods.")
        try:
            self.reset()
            self.step(self.action_space.sample())
            return True
        except Exception as ex:
            if isinstance(ex, NotImplementedError):
                return False
            else:
                raise

    def is_generative(self):
        logging.warning("Checking if Model is \
generative calls sample() method.")
        try:
            self.sample(self.observation_space.sample(),
                        self.action_space.sample())
            return True
        except Exception as ex:
            if isinstance(ex, NotImplementedError):
                return False
            else:
                raise

    @property
    def unwrapped(self):
        return self



import rlberry

_GYM_INSTALLED = True
try:
    import gym
except Exception:
    _GYM_INSTALLED = False


def convert_space_from_gym(gym_space):
    global _GYM_INSTALLED
    assert _GYM_INSTALLED, \
        "gym required by convert_space_from_gym() but not installed."

    if isinstance(gym_space, gym.spaces.Discrete):
        n = gym_space.n
        rlberry_space = rlberry.spaces.Discrete(n)
    elif isinstance(gym_space, gym.spaces.Box):
        assert gym_space.high.ndim == 1, \
            "Conversion from gym.spaces.Box requires high and low to be 1d."
        high = gym_space.high
        low = gym_space.low
        rlberry_space = rlberry.spaces.Box(low, high)

    return rlberry_space



import numpy as np
import pytest
import rlberry.seeding as seeding

from copy import deepcopy
from rlberry.envs.classic_control import MountainCar, Acrobot
from rlberry.envs.finite import Chain
from rlberry.envs.finite import GridWorld
from rlberry.envs.benchmarks.ball_exploration import PBall2D, SimplePBallND
from rlberry.wrappers import Wrapper, RescaleRewardWrapper


_GYM_INSTALLED = True
try:
    import gym
except Exception:
    _GYM_INSTALLED = False


classes = [
    MountainCar,
    GridWorld,
    Chain,
    PBall2D,
    SimplePBallND,
    Acrobot
]


def get_env_trajectory(env, horizon):
    states = []
    ss = env.reset()
    for _ in range(horizon):
        states.append(ss)
        ss, _, _, _ = env.step(env.action_space.sample())
    return states


def compare_trajectories(traj1, traj2):
    for ss1, ss2 in zip(traj1, traj2):
        if not np.array_equal(ss1, ss2):
            return False
    return True


@pytest.mark.parametrize("ModelClass", classes)
def test_wrapper_seeding(ModelClass):

    seeding.set_global_seed(123)
    env1 = Wrapper(ModelClass())

    seeding.set_global_seed(456)
    env2 = Wrapper(ModelClass())

    seeding.set_global_seed(123)
    env3 = Wrapper(ModelClass())

    if deepcopy(env1).is_online():
        traj1 = get_env_trajectory(env1, 500)
        traj2 = get_env_trajectory(env2, 500)
        traj3 = get_env_trajectory(env3, 500)

        assert not compare_trajectories(traj1, traj2)
        assert compare_trajectories(traj1, traj3)


@pytest.mark.parametrize("ModelClass", classes)
def test_rescale_wrapper_seeding(ModelClass):

    seeding.set_global_seed(123)
    env1 = RescaleRewardWrapper(ModelClass(), (0, 1))

    seeding.set_global_seed(456)
    env2 = RescaleRewardWrapper(ModelClass(), (0, 1))

    seeding.set_global_seed(123)
    env3 = RescaleRewardWrapper(ModelClass(), (0, 1))

    if deepcopy(env1).is_online():
        traj1 = get_env_trajectory(env1, 500)
        traj2 = get_env_trajectory(env2, 500)
        traj3 = get_env_trajectory(env3, 500)

        assert not compare_trajectories(traj1, traj2)
        assert compare_trajectories(traj1, traj3)


@pytest.mark.parametrize("ModelClass", classes)
def test_wrapper_copy_reseeding(ModelClass):

    seeding.set_global_seed(123)
    env = Wrapper(ModelClass())

    c_env = deepcopy(env)
    c_env.reseed()

    if deepcopy(env).is_online():
        traj1 = get_env_trajectory(env, 500)
        traj2 = get_env_trajectory(c_env, 500)
        assert not compare_trajectories(traj1, traj2)


@pytest.mark.parametrize("ModelClass", classes)
def test_double_wrapper_copy_reseeding(ModelClass):

    seeding.set_global_seed(123)
    env = Wrapper(Wrapper(ModelClass()))

    c_env = deepcopy(env)
    c_env.reseed()

    if deepcopy(env).is_online():
        traj1 = get_env_trajectory(env, 500)
        traj2 = get_env_trajectory(c_env, 500)
        assert not compare_trajectories(traj1, traj2)


def test_gym_copy_reseeding():
    seeding.set_global_seed(123)
    if _GYM_INSTALLED:
        gym_env = gym.make('Acrobot-v1')
        env = Wrapper(gym_env)

        c_env = deepcopy(env)
        c_env.reseed()

        if deepcopy(env).is_online():
            traj1 = get_env_trajectory(env, 500)
            traj2 = get_env_trajectory(c_env, 500)
            assert not compare_trajectories(traj1, traj2)


def test_gym_copy_reseeding_2():
    seeding.set_global_seed(123)
    if _GYM_INSTALLED:
        gym_env = gym.make('Acrobot-v1')
        # nested wrapping
        env = RescaleRewardWrapper(Wrapper(Wrapper(gym_env)), (0, 1))

        c_env = deepcopy(env)
        c_env.reseed()

        if deepcopy(env).is_online():
            traj1 = get_env_trajectory(env, 500)
            traj2 = get_env_trajectory(c_env, 500)
            assert not compare_trajectories(traj1, traj2)
